#!/usr/bin/env python
# coding=utf-8
import os, re, glob, shutil
import json
import random
import numpy as np
import logging
import logging.handlers
from collections import Counter

try:
    import xlrd, xlwt, openpyxl
except ImportError:
    print('找不到xlrd/xlwt/openpyxl包 不能使用xls相关函数')


def file2list(in_file, strip_nl=True, encoding='U8'):
    with open(in_file, 'r', encoding=encoding) as f:
        lines = [line.strip('\n') if strip_nl else line for line in f]
        print(f'read ok! filename: {in_file}, length: {len(lines)}')
    return lines


def file2txt(in_file, encoding='U8'):
    with open(in_file, 'r', encoding=encoding) as f:
        txt = f.read()
    return txt


# extract: [0,2] or (0,2) or '02'  # assume indices > 9 should not use str
# filter_fn: lambda item: item[2] == 'Y'
def file2items(in_file, strip_nl=True, deli='\t', extract=None, filter_fn=None, encoding='U8'):
    lines = file2list(in_file, strip_nl=strip_nl, encoding=encoding)
    items = [line.split(deli) for line in lines]
    if filter_fn is not None:
        items = list(filter(filter_fn, items))
        print(f'after filter, length: {len(lines)}')
    if extract is not None:
        assert isinstance(extract, (list, tuple, str)), 'invalid extract args'
        items = [[item[int(e)] for e in extract] for item in items]
    return items


# kv_ids: [0,1] or (0,1) or '01'  # assume indices > 9 should not use str
def file2dict(in_file, deli='\t', kv_order='01'):
    items = file2items(in_file, deli=deli)
    assert isinstance(kv_order, (list, tuple, str)) and len(kv_order) == 2, 'invalid kv_order args'
    k_idx = int(kv_order[0])
    v_idx = int(kv_order[1])
    return {item[k_idx]: item[v_idx] for item in items}


# l1,l2,seg_l,l3,l4,l5,seg_l -> [[l1,l2],[l3,l4,l5]]
def file2nestlist(in_file, strip_nl=True, encoding='U8', seg_line=''):
    lst = file2list(in_file, strip_nl=strip_nl, encoding=encoding)
    out_lst_lst = []
    out_lst = []
    for line in lst:
        if line == seg_line:
            out_lst_lst.append(out_lst)
            out_lst = []
            continue
        out_lst.append(line)
    if out_lst:
        out_lst_lst.append(out_lst)
    return out_lst_lst


# [l1,l2,seg,l3,l4,l5] -> [[l1,l2],[l3,l4,l5]]
def seg_list(lst, is_seg_fn=None):
    if is_seg_fn is None:
        is_seg_fn = lambda e: e == ''
    nest_lst = [[]]
    for e in lst:
        if is_seg_fn(ele):
            nest_lst.append([])
            continue
        nest_lst[-1].append(ele)
    return nest_lst


def obj2json(obj, out_file, indent=4):
    def np2py(obj):
        # np格式不支持json序列号,故转化为python数据类型
        if isinstance(obj, (int, float, bool, str)):
            return obj
        if isinstance(obj, dict):
            for k in obj:
                obj[k] = np2py(obj[k])
            return obj
        elif isinstance(obj, (list, tuple)):
            for i in range(len(obj)):
                obj[i] = np2py(obj[i])
            return obj
        elif isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return obj

    obj = np2py(obj)
    json.dump(obj, open(out_file, 'w', encoding='U8'),
              ensure_ascii=False, indent=indent)
    print(f'save json file ok! {out_file}')


def list_dir_and_file(path):
    lsdir = os.listdir(path)
    dirs = [d for d in lsdir if os.path.isdir(os.path.join(path, d))]
    files = [f for f in lsdir if os.path.isfile(os.path.join(path, f))]
    return dirs, files


def list2file(lines, out_file, add_nl=True, deli='\t'):
    # 兼容
    if isinstance(lines, str):
        lines, out_file = out_file, lines
    assert len(lines) > 0, 'lines must be not None'
    with open(out_file, 'w', encoding='U8') as f:
        if isinstance(lines[0], (list, tuple)):
            lines = [deli.join(map(str, item)) for item in lines]
        # other: str, int, float, bool, obj will use f'{} to strify
        out_list = [f'{line}\n' if add_nl else f'{line}' for line in lines]
        f.writelines(out_list)
        print(f'save ok! filename: {out_file}, length: {len(out_list)}')


def freqs(lst):
    c = Counter(lst)
    return c.most_common()


def list2stats(in_lst, out_file=None):
    stats = freqs(in_lst)
    print(*stats, sep='\n')
    if out_file is not None:
        list2file(stats, out_file, deli='\t')


# sheet_ids: [0,1,3] default read all sheet
def xls2items(in_xls, start_row=1, sheet_ids=None):
    items = []
    xls = xlrd.open_workbook(in_xls)
    sheet_ids = list(range(xls.nsheets)) if sheet_ids is None else sheet_ids
    for sheet_id in sheet_ids:
        sheet = xls.sheet_by_index(sheet_id)
        nrows, ncols = sheet.nrows, sheet.ncols
        print(f'reading... sheet_id:{sheet_id} sheet_name:{sheet.name} rows:{nrows} cols:{ncols}')
        for i in range(start_row, nrows):
            items.append([sheet.cell_value(i, j) for j in range(ncols)])
    return items


# this only support old xls (nrows<65537)
def items2xls_old(items, out_xls=None, sheet_name=None, header=None, workbook=None, max_row_per_sheet=65537):
    workbook = xlwt.Workbook(encoding='utf-8') if workbook is None else workbook
    sheet_name = '1' if sheet_name is None else sheet_name
    num_sheet = 1
    worksheet = workbook.add_sheet(f'{sheet_name}_{num_sheet}')  # 创建一个sheet
    if header is not None:
        for j in range(len(header)):
            worksheet.write(0, j, header[j])
    row_ptr = 1 if header is not None else 0
    for item in items:
        if row_ptr + 1 > max_row_per_sheet:
            num_sheet += 1
            worksheet = workbook.add_sheet(f'{sheet_name}_{num_sheet}')
            if header is not None:
                for j in range(len(header)):
                    worksheet.write(0, j, header[j])
            row_ptr = 1 if header is not None else 0

        for j in range(len(item)):
            worksheet.write(row_ptr, j, item[j])
        row_ptr += 1
    if out_xls is not None:  # 如果为None表明调用者其实还有新的items要加到新的sheet要中，只想要返回的workbook对象
        workbook.save(out_xls)
        print(f'save ok! xlsname: {out_xls}, num_sheet: {num_sheet}')
    return workbook


def items2xls(items, out_xls=None, sheet_name=None, header=None, workbook=None, max_row_per_sheet=65537):
    if workbook is None:
        workbook = openpyxl.Workbook()  # create new workbook instance
        active_worksheet = workbook.active
        workbook.remove(active_worksheet)
    if sheet_name is None:
        sheet_name = 'sheet'
    num_sheet = 1
    worksheet = workbook.create_sheet(f'{sheet_name}_{num_sheet}' if len(items) > max_row_per_sheet else f'{sheet_name}')  # 创建一个sheet
    if header is not None:
        for j in range(len(header)):
            worksheet.cell(0 + 1, j + 1, header[j])  # cell x y 从1开始
    row_ptr = 1 if header is not None else 0
    for item in items:
        if row_ptr + 1 > max_row_per_sheet:
            num_sheet += 1
            worksheet = workbook.create_sheet(f'{sheet_name}_{num_sheet}')
            if header is not None:
                for j in range(len(header)):
                    worksheet.cell(0 + 1, j + 1, header[j])
            row_ptr = 1 if header is not None else 0

        for j in range(len(item)):
            worksheet.cell(row_ptr + 1, j + 1, item[j])
        row_ptr += 1
    if out_xls is not None:  # 如果为None表明调用者其实还有新的items要加到新的sheet表中，只想要返回的workbook对象
        workbook.save(out_xls)
        print(f'save ok! xlsname: {out_xls}, num_sheet: {num_sheet}')
    return workbook


def merge_file(file_list, out_file, shuffle=False):
    assert isinstance(file_list, (list, tuple))
    ret_lines = []
    for i, file in enumerate(file_list):
        lines = file2list(file, strip_nl=False)
        print(f'已读取第{i}个文件:{file}\t行数{len(lines)}')
        ret_lines.extend(lines)
    if shuffle:
        random.shuffle(ret_lines)
    list2file(out_file, ret_lines, add_nl=False)


def merge_file_by_pattern(pattern, out_file, shuffle=False):
    file_list = glob.glob(pattern)
    merge_file(file_list, out_file, shuffle)


# ratio: [18,1,1] or '18:1:1'
# num: 1000
def split_file(file_or_lines, num=None, ratio=None, files=None, shuffle=True, seed=None):
    assert num or ratio, 'invalid args: at least use num or ratio'
    if type(file_or_lines) == str:
        lines = file2list(file_or_lines, strip_nl=False)
    else:
        assert isinstance(file_or_lines, (list, tuple)), 'invalid args file_or_lines'
        lines = file_or_lines
    length = len(lines)
    if shuffle:
        if seed is not None:
            random.seed(seed)
        random.shuffle(lines)
    if num:
        assert num < length, f'invalid args num: num:{num} should < filelen: {length}'
        lines1 = lines[:num]
        lines2 = lines[num:]
        if files:
            assert len(files) == 2
            list2file(files[0], lines1, add_nl=False)
            list2file(files[1], lines2, add_nl=False)
        return lines1, lines2
    if ratio:  # [6,2,2]
        if isinstance(ratio, str):
            ratio = list(map(int, ratio.split(':')))
        cumsum_ratio = np.cumsum(ratio)  # [6,8,10]
        sum_ratio = cumsum_ratio[-1]  # 10
        assert sum_ratio <= length, f'invalid args ratio: ratio:{ratio} should <= filelen: {length}'
        indices = [length * r // sum_ratio for r in cumsum_ratio]  # [6,8,11] if length=11
        indices = [0] + indices
        split_lines_lst = []
        for i in range(len(indices) - 1):
            split_lines_lst.append(lines[indices[i]:indices[i + 1]])
        if files:
            assert len(files) == len(split_lines_lst)
            for i, lines in enumerate(split_lines_lst):
                list2file(files[i], lines, add_nl=False)
        return split_lines_lst


def delete_file(file_or_dir, verbose=True):
    if os.path.exists(file_or_dir):
        if os.path.isfile(file_or_dir):  # 文件file
            os.remove(file_or_dir)
            if verbose:
                print(f'delete ok! file: {file_or_dir}')
            return True
        else:  # 目录dir
            for file_lst in os.walk(file_or_dir):
                for name in file_lst[2]:
                    os.remove(os.path.join(file_lst[0], name))
            shutil.rmtree(file_or_dir)
            if verbose:
                print(f'delete ok! dir: {file_or_dir}')
            return True
    else:
        print(f'delete false! file/dir not exists: {file_or_dir}')
        return False


def find_duplicates(in_lst):
    # duplicates = []
    # seen = set()
    # for item in in_lst:
    #     if item not in seen:
    #         seen.add(item)
    #     else:
    #         duplicates.append(item)
    # return duplicates
    c = Counter(in_lst)
    return [k for k, v in c.items() if v > 1]


def remove_duplicates(in_file, out_file=None, keep_sort=True):
    if not out_file:
        out_file = in_file
    lines = file2list(in_file)

    if keep_sort:
        out_lines = []
        tmp_set = set()
        for line in lines:
            if line not in tmp_set:
                out_lines.append(line)
                tmp_set.add(line)
    else:
        out_lines = list(set(lines))

    list2file(out_file, out_lines)


def set_items(items, keep_order=False):
    items = [tuple(item) for item in items]  # need to be hashable i.e. tuple
    ret = []
    if keep_order:
        seen = set()
        for item in items:
            if item not in seen:
                seen.append(item)
                ret.append(list(item))
        return ret
    if not keep_order:
        ret = list(map(list, set(items)))
        return ret


def sort_items(items, sort_order):
    if isinstance(sort_order, str):
        sort_order = map(int, sort_order)
    for idx in reversed(sort_order):
        items.sort(key=lambda item: item[idx])
    return items


def check_overlap(list1, list2, verbose=False):
    set1 = set(list1)
    set2 = set(list2)
    count1 = Counter(list1)
    dupli1 = [k for k, v in count1.items() if v > 1]
    count2 = Counter(list2)
    dupli2 = [k for k, v in count2.items() if v > 1]

    print(f'原始长度:{len(list1)}\t去重长度{len(set1)}\t重复项{dupli1}')
    print(f'原始长度:{len(list2)}\t去重长度{len(set2)}\t重复项{dupli2}')

    union = sorted(set1 & set2)  # 变为list
    print(f'一样的数量: {len(union)}')
    if verbose or len(union) <= 30:
        print(*union, sep='\n', end='\n\n')
    else:
        print(*union[:30], sep='\n', end=f'\n ..more(total:{len(union)})\n\n')

    a = sorted(set1 - set2)  # 变为list
    print(f'前者多了: {len(a)}')
    if verbose or len(a) <= 30:
        print(*a, sep='\n', end='\n\n')
    else:
        print(*a[:30], sep='\n', end=f'\n ..more(total:{len(a)})\n\n')

    b = sorted(set2 - set1)  # 变为list
    print(f'后者多了: {len(b)}')
    if verbose or len(b) <= 30:
        print(*b, sep='\n', end='\n\n')
    else:
        print(*b[:30], sep='\n', end=f'\n ..more(total:{len(a)})\n\n')


def print_len(files):
    if not isinstance(files, list):
        files = [files]
    len_lst = []
    for file in files:
        with open(file, 'r', encoding='U8') as f:
            len_lst.append(len(f.readlines()))
    print(len_lst, f'总和: {sum(len_lst)}')


def f(object):
    """ 格式化 """
    if 'numpy' in str(type(object)) and len(object.shape) == 1:
        object = object.tolist()
    if isinstance(object, (list, tuple)):
        if len(object) == 0:
            return ''
        if isinstance(object[0], (int, float)):
            ret = list(map(lambda e: f'{e:.2f}', object))
            return str(ret)
    return str(object)


class Dict(dict):
    __setattr__ = dict.__setitem__
    __getattr__ = dict.__getitem__


def dict2obj(dct):
    if not isinstance(dct, dict):
        return dct
    inst = Dict()
    for k, v in dct.items():
        inst[k] = dict2obj(v)
    return inst


class ImmutableDict(dict):
    def __hash__(self):
        return id(self)

    def _immutable(self, *args, **kws):
        raise TypeError('object is immutable')

    __setitem__ = _immutable
    __delitem__ = _immutable
    clear = _immutable
    update = _immutable
    setdefault = _immutable
    pop = _immutable
    popitem = _immutable


class Any2Id():
    # fix 是否固定住字典不在增加
    def __init__(self, use_line_no=False, exist_dict=None, fix=True, counter=None):
        self.fix = fix
        self.use_line_no = use_line_no
        self.counter = Counter() if not counter else counter
        self.any2id = {}  # 内部核心dict
        if exist_dict is not None:
            self.any2id.update(exist_dict)

        # for method_name in dict.__dict__:  # 除了下面显式定义外保证能以dict的各种方法操作Any2id
        #     setattr(self, method_name, getattr(self.any2id, method_name))

    def keys(self):
        return self.any2id.keys()

    def values(self):
        return self.any2id.values()

    def items(self):
        return self.any2id.items()

    def pop(self, key):
        return self.any2id.pop(key)

    def __getitem__(self, item):
        return self.any2id.__getitem__(item)

    def __setitem__(self, key, value):
        self.any2id.__setitem__(key, value)

    def __len__(self):
        return self.any2id.__len__()

    def __iter__(self):
        return self.any2id.__iter__()

    def __str__(self):
        return self.any2id.__str__()

    def set_fix(self, fix):
        self.fix = fix

    def get(self, key, default=None, add=False):
        if not add:
            return self.any2id.get(key, default)
        else:
            # new_id = len(self.any2id)
            new_id = max(self.any2id.values()) + 1
            self.any2id[key] = new_id
            return new_id

    def get_reverse(self):
        return self.__class__(exist_dict={v: k for k, v in self.any2id.items()})

    def save(self, file, use_line_no=None, deli='\t'):
        if use_line_no is None:
            use_line_no = self.use_line_no
        items = sorted(self.any2id.items(), key=lambda e: e[1])
        out_items = [item[0] for item in items] if use_line_no else items
        list2file(out_items, file, deli=deli)
        print(f'词表文件生成成功: {file} {items[:5]}...')

    def load(self, file, use_line_no=None, deli='\t'):
        if use_line_no is None:
            use_line_no = self.use_line_no
        items = file2items(file, deli=deli)
        if use_line_no or len(items[0]) == 1:
            self.any2id = {item[0]: i for i, item in enumerate(items)}
        else:
            self.any2id = {item[0]: int(item[1]) for item in items}

    def to_count(self, any_lst):
        self.counter.update(any_lst)  # 维护一个计数器

    def reset_counter(self):
        self.counter = Counter()

    def rebuild_by_counter(self, restrict=None, min_freq=None, max_vocab_size=None):
        if not restrict:
            restrict = ['<pad>', '<unk>', '<eos>']
        freqs = self.counter.most_common()
        tokens_lst = restrict[:]
        curr_vocab_size = len(tokens_lst)
        for token, cnt in freqs:
            if min_freq and cnt < min_freq:
                break
            if max_vocab_size and curr_vocab_size >= max_vocab_size:
                break
            tokens_lst.append(token)
            curr_vocab_size += 1
        self.any2id = {token: i for i, token in enumerate(tokens_lst)}

    @classmethod
    def from_file(cls, file, use_line_no=False, deli='\t'):
        inst = cls(use_line_no=use_line_no)
        if os.path.exists(file):
            inst.load(file, deli=deli)
        else:
            # will return inst with empty any2id , e.g. boolean(inst) or len(inst) will return False
            print(f'vocab file: {file} not found, need to build and save later')
        return inst


def pad_sequences(sequences, maxlen=None, dtype='int32', padding='pre', truncating='pre', value=0.):
    if not hasattr(sequences, '__len__') or not all([hasattr(x, '__len__') for x in sequences]):
        raise ValueError(f'sequences invalid: {sequences}')

    len_lst = [len(x) for x in sequences]

    if len(set(len_lst)) == 1:  # 长度均相等
        ret = np.array(sequences)
        if maxlen is not None:
            ret = ret[:,-maxlen:] if truncating=='pre' else ret[:,:maxlen]
        return ret

    num_samples = len(sequences)
    if maxlen is None:
        maxlen = np.max(len_lst)

    sample_shape = tuple()
    for s in sequences:
        if len(s) > 0:
            sample_shape = np.asarray(s).shape[1:]
            break

    is_dtype_str = np.issubdtype(dtype, np.str_) or np.issubdtype(dtype, np.unicode_)
    if isinstance(value, str) and dtype != object and not is_dtype_str:
        raise ValueError("`dtype` {} is not compatible with `value`'s type: {}\n"
                         "You should set `dtype=object` for variable length strings."
                         .format(dtype, type(value)))

    x = np.full((num_samples, maxlen) + sample_shape, value, dtype=dtype)
    for idx, s in enumerate(sequences):
        if not len(s):
            continue  # empty list/array was found
        if truncating == 'pre':
            trunc = s[-maxlen:]
        elif truncating == 'post':
            trunc = s[:maxlen]
        else:
            raise ValueError(f'Truncating type not support')

        # check `trunc` has expected shape
        trunc = np.asarray(trunc, dtype=dtype)
        if trunc.shape[1:] != sample_shape:
            raise ValueError('Shape of sample %s of sequence at position %s '
                             'is different from expected shape %s' %
                             (trunc.shape[1:], idx, sample_shape))

        if padding == 'post':
            x[idx, :len(trunc)] = trunc
        elif padding == 'pre':
            x[idx, -len(trunc):] = trunc
        else:
            raise ValueError('Padding type not support')
    return x


def get_file_logger(logger_name, log_file='./qiznlp.log', level='DEBUG'):
    level = {
        'ERROR': logging.ERROR,
        'INFO': logging.INFO,
        'DEBUG': logging.DEBUG,
        'NOTSET': logging.NOTSET,
    }.get(level, logging.DEBUG)

    logger = logging.getLogger(logger_name)
    logger.setLevel(level)

    fh = logging.handlers.TimedRotatingFileHandler(log_file, when="D", interval=1, backupCount=7, encoding="utf-8")
    fh.suffix = "%Y-%m-%d.log"  # 设置后缀名称，跟strftime的格式一样
    fh.extMatch = re.compile(r"^\d{4}-\d{2}-\d{2}.log$")  # _\d{2}-\d{2}
    fh.setLevel(level)
    fh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
    logger.addHandler(fh)
    logger.propagate = False  # 取消传递，只有这样才不会传到logger_root然后又在控制台也打印一遍

    return logger

def suppress_tf_warning(tf):
    import warnings
    warnings.filterwarnings('ignore')
    tf.logging.set_verbosity(tf.logging.ERROR)

if __name__ == '__main__':
    check_overlap([1, 2, 3, 4, 5, 6], [4, 5, 6, 7, 8, 9, 0])
    exit(0)

    a = Any2Id()
    data = [random.randint(1, 100) for _ in range(10000)]
    for ele in data:
        a.to_count([ele])

    a.rebuild_by_counter(['<pad>', '<unk>'])



#!/usr/bin/env python
# coding=utf-8
import tensorflow as tf
from tensorflow.python.util import nest

INF = 1e7
# INF = float('inf')

# debug
TF_PRINT = False


def tf_print(tensor, message=''):
    if not TF_PRINT:
        return tensor
    else:
        return tf.Print(tensor, [tf.shape(tensor), tensor], message='Debug:%s' % message, summarize=40)


def log_prob_from_logits(logits, axis=-1):
    return logits - tf.reduce_logsumexp(logits, axis=axis, keepdims=True)


def shape_list(x):
    """ return list of dims, statically where possible """
    x = tf.convert_to_tensor(x)
    # if unknown rank, return dynamic shape 如果秩都不知道
    if x.get_shape().dims is None:
        return tf.shape(x)
    static = x.get_shape().as_list()
    shape = tf.shape(x)
    ret = []
    for i in range(len(static)):
        dim = shape[i] if static[i] is None else static[i]
        ret.append(dim)
    return ret


def assign_value_to_tensor(tensor, idx, value):
    """ tensor 1-D
        idx scalar
        value scalar
        ret 1-D
    """
    dtype = tensor.dtype
    length = tf.shape(tensor)[0]

    weight_one_target_idx = tf.one_hot(idx, length, on_value=1., off_value=0.)
    weight_zero_target_idx = 1. - weight_one_target_idx

    weight_one_target_idx = tf.cast(weight_one_target_idx, dtype=dtype)
    weight_zero_target_idx = tf.catt(weight_zero_target_idx, dtype=dtype)
    ret = tensor * weight_zero_target_idx + value * weight_one_target_idx
    return ret


def multinomial_without_repeat(logits, num_samples, top_k, uniform=False, eos_id=1):
    """ logits 2-D
        num_samples scalar
        topk scalar
        uniform bool
        ret [batch,num_samples]
    """
    batch_size = tf.shape(logits)[0]
    init_i = tf.constant(0)
    init_logits = logits
    init_ret = tf.zeros([batch_size, 0], dytpe=tf.int32)

    # 先把eos设置为不可能(-INF)
    init_logits = tf.map_fn(lambda x: assign_value_to_tensor(x, eos_id, value=-INF), init_logits)

    init_topk_logits, init_topk_ids = tf.nn.top_k(init_logits, k=top_k, sorted=False)  # [batch,topk]
    if uniform:
        init_topk_logits = tf.zeros_like(init_topk_logits)

    def loop_condition(i, *_):
        # i < N and i < topk
        return tf.logical_and(tf.less(i, num_samples), tf.less(i, top_k))

    def loop_body(i, logits, ret):
        sample_idx = tf.multinomial(logits, 1, output_dtype=tf.int32)  # [batch,1]
        # 计算原始vocab_vector里面的idx
        batch_pos = tf.reshape(tf.range(batch_size), [batch_size, 1])  # [batch,1]
        ids_pos = tf.stack([batch_pos, sample_idx], axis=-1)  # [batch,1,2]
        ori_sample_idx = tf.gather_nd(init_topk_ids, ids_pos)  # [batch,1]

        ret = tf.concat([ret, ori_sample_idx], axis=-1)
        sample_idx = tf.squeenze(sample_idx, axis=-1)  # [batch]
        logits = tf.map_fn(lambda x: (assign_value_to_tensor(x[0], x[1], value=-INF), x[1]), (logits, sample_idx))  # tuple(logits,tmp)
        logits = logits[0]
        i += 1
        return i, logits, ret

    final_i, final_logits, final_ret = tf.while_loop(loop_condition,
                                                     loop_body,
                                                     [init_i, init_topk_logits, init_ret],
                                                     shape_invariants=[tf.TensorShape([]),
                                                                       tf.TensorShape([None, None]),
                                                                       tf.TensorShape([None, None])]
                                                     )
    return final_ret


def merge_pre_n_dim(tensor, n=2):
    """ n=2 [a,b,...] -> [a*b,...] """
    shape = shape_list(tensor)
    for i in range(1, n):
        shape[0] *= shape[i]
    shape = shape[:1] + shape[n:]
    return tf.reshape(tensor, shape)


def unmerge_first_dim(tensor, sizes):
    """ sizes=[a,b] [a*b,...] -> [a,b,...] """
    shape = shape_list(tensor)
    shape = sizes + shape[1:]
    return tf.reshape(tensor, shape)


def expand_and_tile_to_ndims(tensor, sizes):
    """ sizes=[beam] [a,...] -> [a,beam,...] """
    n_dims = len(sizes)
    for _ in range(n_dims):
        tensor = tf.expand_dims(tensor, axis=1)
    tile_dims = [1] * tensor.shape.ndims
    for i in range(n_dims):
        tile_dims[i + 1] = sizes[i]
    return tf.tile(tensor, tile_dims)


def get_state_shape_invariants(tensor):
    """ return the shape of tensor but set middle dims to None """
    shape = tensor.shape.as_list()
    for i in range(1, len(shape) - 1):
        shape[i] = None
    return tf.TensorShape(shape)


def compute_batch_indices(batch_size, beam_size):
    batch_pos = tf.range(batch_size * beam_size) // beam_size
    batch_pos = tf.reshape(batch_pos, [batch_size, beam_size])
    return batch_pos


def compute_batch_and_group_indices(batch_size, group_size, beam_size):
    batch_pos = tf.range(batch_size * group_size * beam_size) // (group_size * beam_size)
    batch_pos = tf.reshape(batch_pos, [batch_size, group_size, beam_size])

    group_pos = tf.range(group_size * beam_size) // beam_size
    group_pos = tf.reshape(group_pos, [group_size, beam_size])  # [group,beam]
    group_pos = tf.expand_dims(group_pos, axis=0)
    group_pos = tf.tile(group_pos, [batch_size, 1, 1])  # [batch,group,beam]

    return tf.stack([batch_pos, group_pos], axis=3)  # [batch,group,beam,2


def beam_search(symbols_to_logits_fn,
                initial_ids,  # int32
                beam_size,
                max_decode_len,
                vocab_size,
                alpha=0,
                states=None,
                eos_id=1,
                stop_early=True,
                num_group=1,
                gamma=0,
                stop_nums=None,
                top_k=30):
    """ """
    group_size = num_group  # 换个名字
    beam_size = beam_size // group_size
    use_group = tf.not_equal(group_size, 1)  # 是否分组flag

    batch_size = shape_list(initial_ids)[0]
    # initial log probs
    init_log_probs = tf.concat([tf.constant([0.]), tf.tile(tf.constant([-INF]), [beam_size - 1])], axis=0)  # [beam]
    init_log_probs = tf.expand_dims(tf.expand_dims(init_log_probs, axis=0), axis=0)  # [1,1,beam]
    init_log_probs = tf.tile(init_log_probs, [batch_size, group_size, 1])  # [batch,group,beam]

    # initial seq
    init_seq = expand_and_tile_to_ndims(initial_ids, [group_size, beam_size])  # [batch,group,beam]
    init_seq = tf.expand_dims(init_seq, axis=3)  # [batch,group,beam,1] 使用pad当做eos

    # initial state
    if states:  # [batch,...]
        states = nest.map_structure(lambda t: expand_and_tile_to_ndims(t, [group_size, beam_size]), states)  # [batch,group,beam,...]

    finished_flags = tf.fill([batch_size, group_size, beam_size], False)  # [batch,group,beam]

    def loop_body(i, seq, log_probs, finished_flags, states):
        """ """
        # flat the seq
        flat_ids = tf.reshape(seq, [batch_size * group_size * beam_size, -1])  # [batch*group*beam,length]
        if states:
            flat_states = nest.map_structure(lambda t: merge_pre_n_dim(t, n=3), states)
            flat_logits, flat_states = symbols_to_logits_fn(flat_ids, i, flat_states)
            states = nest.map_structure(lambda t: unmerge_first_dim(t, [batch_size, group_size, beam_size]), flat_states)
        else:
            flat_logits = symbols_to_logits_fn(flat_ids, i, states)

        logits = tf.reshape(flat_logits, [batch_size, group_size, beam_size, -1])  # [batch,group,beam,vocab]

        # calc log-prob
        candidate_log_probs = log_prob_from_logits(logits)

        def process_diversity():
            """ 多样性鼓励因子，1个group中的多样性 """
            _candidate_log_probs = tf.reshape(candidate_log_probs, [-1, vocab_size])
            _, topk_candidate_log_probs_ids = tf.nn.top_k(_candidate_log_probs, k=vocab_size)
            diversity_rank = tf.map_fn(tf.invert_permutation, topk_candidate_log_probs_ids) + 1
            diversity_rank = tf.to_float(diversity_rank) * -1. * gamma
            _candidate_log_probs += diversity_rank
            _candidate_log_probs = tf.reshape(_candidate_log_probs, [batch_size, group_size, beam_size, vocab_size])
            return _candidate_log_probs

        # 如果部分组则应从第一步开始加上多样性因子，如果分组应从第二步开始
        start_step = tf.cond(use_group, lambda: 1, lambda: 0)
        candidate_log_probs = tf.cond(tf.greater(i, start_step), process_diversity, lambda: candidate_log_probs)

        # 使结束的句子下一步强制选择eos, 通过调整下一步的candidate_log_probs
        eos_one_hot_vec = tf.one_hot(eos_id, vocab_size, on_value=1., off_value=0.)  # [vocab]
        vocab_mask = -1 * eos_one_hot_vec + INF * (1 - eos_one_hot_vec)  # [vocab]
        vocab_mask = tf.expand_dims(vocab_mask, axis=0)  # [1,vocab]
        flat_finished_flags = tf.reshape(tf.to_float(finished_flags), [-1, 1])  # [batch,group,beam,1]

        mask = tf.matmul(flat_finished_flags, vocab_mask)  # [batch,group,beam,vocab]
        mask += 1
        mask = tf.reshape(mask, [batch_size, group_size, beam_size, vocab_size])

        candidate_log_probs *= mask

        # 加上前面已计算分数的log_probs
        log_probs = candidate_log_probs + tf.expand_dims(log_probs, axis=3)  # [batch,group,beam,vocab]

        # 平展beam和vocab用以组内全局比较
        flat_log_probs = tf.reshape(log_probs, [-1, group_size, beam_size * vocab_size])  # [batch,group,beam*vocab]

        def group_first_step_sample_from_logits():
            """ 根据logits进行首字采样，必须分组才能使用本功能"""
            batch_logits = tf.reshape(flat_logits, [batch_size, group_size, beam_size, -1])  # [batch,group,beam,vocab]
            batch_logits = batch_logits[:, 0, 0, :]  # [batch,vocab]  # 由于首字每个beam里的所有logits一样
            sample_idx = multinomial_without_repeat(batch_logits, group_size, top_k=top_k, uniform=False)  # [batch,group]
            # idx 变换为[batch*group*beam*1] 利于下一步取出对应log_prob
            sample_idx = tf.expand_dims(sample_idx, axis=-1)  # [batch,group,1]
            sample_idx = tf.tile(sample_idx, [1, 1, beam_size])  # [batch,group,beam]
            sample_idx = tf.reshape(sample_idx, [-1, 1])  # [batch*group*beam,1]
            # 取出对应log_prob分数
            flat_log_prob = log_prob_from_logits(flat_logits)  # [batch*group*beam,vocab]
            bgbid = tf.expand_dims(tf.range(batch_size * group_size * beam_size), axis=-1)  # [batch*group*beam,1]
            bgbid = tf.concat([bgbid, sample_idx], axis=-1)  # [batch*group*beam,2]
            sample_log_prob = tf.gather_nd(flat_log_prob, bgbid)  # [batch*group*beam,vocab]
            # 变换为正式形式
            sample_idx = tf.reshape(sample_idx, [batch_size, group_size, beam_size])  # [batch,group,beam]
            sample_log_prob = tf.reshape(sample_log_prob, [batch_size, group_size, beam_size])  # [batch,group,beam]
            # 每组mask除第一个以外的分数，适配分组beam_search的第二步
            log_prob_mask = tf.ones_like(sample_log_prob) * -INF
            sample_log_prob = tf.concat([sample_log_prob[:, :, :1], log_prob_mask[:, :, 1:]], axis=-1)
            sample_log_prob = tf_print(sample_log_prob, 'sample_log_prob')
            sample_idx = tf_print(sample_idx, 'sample_idx')
            return sample_log_prob, sample_idx

        def group_first_step_argmax_from_logprob():
            """ 需每组的开头第一个字是相同的，组间不同
                e.g.
                batch=2 group=3 beam=2
                ret
                ids [[a,a],[b,b],[c,c]]
                scores [[a,-inf],[b,-inf],[c,-inf]]
            """
            _topk_log_probs, _topk_ids = tf.nn.top_k(flat_log_probs, k=group_size)  # [batch,group,group]
            _topk_ids = _topk_ids[:, 0, :]  # [batch,group]
            _topk_log_probs = _topk_log_probs[:, 0, :]  # [batch,group]

            _topk_ids = tf.expand_dims(_topk_ids, axis=-1)  # [batch,group,1]
            _topk_ids = tf.tile(_topk_ids, [1, 1, beam_size])  # [batch,group,beam]

            _topk_log_probs = tf.expand_dims(_topk_log_probs, axis=-1)  # [batch,group,1]
            # 每组mask掉除第一个以外的分数,适配分组beam_search的第二步
            log_prob_mask = tf.ones([batch_size, group_size, beam_size]) * -INF  # [batch,group,beam]
            _topk_log_probs = tf.concat([_topk_log_probs, log_prob_mask[:, :, 1:]], axis=2)

            return _topk_log_probs, _topk_ids

        def other_step():
            _topk_log_probs, _topk_ids = tf.nn.top_k(flat_log_probs, k=beam_size)  # [batch,group,beam]
            return _topk_log_probs, _topk_ids

        # 如果不分组，就不执行first_step了
        use_group_first_step = tf.cond(use_group, lambda: 0, lambda: -1)
        topk_log_probs, topk_ids = tf.cond(tf.equal(i, use_group_first_step), group_first_step_argmax_from_logprob, other_step)

        next_log_probs = topk_log_probs  # [batch,group,beam]

        # calc what group the top probs are in
        topk_seq_idx = topk_ids // vocab_size  # [batch,group,beam]
        topk_ids %= vocab_size  # unflatten the ids [batch,group,beam]

        batch_and_group_pos = compute_batch_and_group_indices(batch_size, group_size, beam_size)  # [batch,group,beam,2]
        topk_coord = tf.concat([batch_and_group_pos, tf.expand_dims(topk_seq_idx, axis=-1)], axis=3)  # [batch,group,beam,3]

        finished_flag = tf.equal(topk_ids, eos_id)
        if states:
            states = nest.map_structure(lambda s: tf.gather_nd(s, topk_coord), states)

        next_seq = tf.gather_nd(seq, topk_coord)
        next_seq = tf.concat([next_seq, tf.expand_dims(topk_ids, axis=3)], axis=3)

        return i + 1, next_seq, next_log_probs, finished_flag, states

    def loop_condition(i, unused_seq, unused_log_probs, finished_flags, unused_states):
        not_all_finish = tf.logical_not(tf.reduce_all(finished_flags))
        less_than_maxlength = tf.less(i, max_decode_len)
        return tf.logical_and(not_all_finish, less_than_maxlength)

    # start loop
    _, seq, log_probs, finished_flags, _ = tf.while_loop(
        loop_condition,
        loop_body,
        [tf.constant(0), init_seq, init_log_probs, finished_flags, states],
        shape_invariants=[
            tf.TensorShape([]),
            tf.TensorShape([None, None, None, None]),
            tf.TensorShape([None, None, None]),
            tf.TensorShape([None, None, None]),
            # finished_flags.get_shape(),
            nest.map_structure(get_state_shape_invariants, states)]
    )

    # 合并group和beam
    seq = tf.reshape(seq, [batch_size, group_size * beam_size, -1])  # [batch,beam*group,len]
    log_probs = tf.reshape(log_probs, [batch_size, group_size * beam_size])  # [batch,beam*group]

    seq = tf_print(seq, 'seq')
    log_probs = tf_print(log_probs, 'log_probs')

    return seq[:, :, 1:], log_probs  # 去除起始符


def sample_with_temperature(logits, temperature):
    """ 0.0:argmax 1.0:sampling >1.0:random """
    # logits [batch,length,vocab]
    # ret [batch,length]
    if temperature == 0.0:
        return tf.argmax(logits, axis=-1)
    else:
        assert temperature > 0.0
        logits_shape = shape_list(logits)
        reshape_logits = tf.reshape(logits, [-1, logits_shape[-1]]) / temperature
        choices = tf.multinomial(reshape_logits, 1)  # 仅采样1个 该方式只支持2-D
        choices = tf.reshape(choices, logits_shape[:-1])
        return choices


def greedy_search(symbols_to_logits_fn,
                  initial_ids,  # int32
                  max_decode_len,
                  cache=None,
                  eos_id=1,
                  sampling_method='argmax',
                  sampling_temp=0.):
    """  """
    # initial_ids: [batch] tf.zeros([batch_size], dtype=tf.int32) 
    batch_size = shape_list(initial_ids)[0]
    decoded_ids = tf.expand_dims(initial_ids, 1)  # [batch,1]
    initial_log_prob = tf.zeros([batch_size], dtype=tf.float32)
    hit_eos = tf.fill([batch_size], False)

    def inner_loop(i, hit_eos, decoded_ids, cache, log_prob):
        """ one step for greedy decoding """
        logits, cache = symbols_to_logits_fn(decoded_ids, i, cache)
        log_probs = log_prob_from_logits(logits)
        temperature = 0.0 if sampling_method == 'argmax' else sampling_temp
        next_id = tf.cast(sample_with_temperature(logits, temperature), dtype=tf.int32)  # [batch]
        hit_eos |= tf.equal(next_id, eos_id)

        log_prob_indices = tf.stack([tf.range(tf.to_int32(batch_size)), next_id], axis=1)
        log_prob += tf.gather_nd(log_probs, log_prob_indices)

        decoded_ids = tf.concat([decoded_ids, tf.expand_dims(next_id, axis=1)], axis=1)
        return i + 1, hit_eos, decoded_ids, cache, log_prob

    def is_not_finished(i, hit_eos, *_):
        finished = (i >= max_decode_len)
        finished |= tf.reduce_all(hit_eos)
        return tf.logical_not(finished)

    _, _, decoded_ids, _, log_prob = tf.while_loop(
        is_not_finished,
        inner_loop,
        [tf.constant(0), hit_eos, decoded_ids, cache, initial_log_prob],
        shape_invariants=[
            tf.TensorShape([]),
            tf.TensorShape([None]),
            tf.TensorShape([None, None]),
            nest.map_structure(get_state_shape_invariants, cache),
            tf.TensorShape([None]),
        ])

    decoded_ids = tf.expand_dims(decoded_ids, 1)[:, :, 1:]  # 去除齐师傅
    log_prob = tf.expand_dims(log_prob, 1)
    return decoded_ids, log_prob  # [batch,1,len], [batch,1]


def symbols_to_logits_fn_readme():
    """ 
        输入：ids, i, cache
            ids: [batch,len] 一般直接取最末尾一个进行计算
        输出：logits, cache
            logits: [batch,h]  len维度没有
    """
    pass



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions and classes related to optimization (weight updates)."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import tensorflow as tf


def create_optimizer(loss, init_lr, num_train_steps, num_warmup_steps, use_tpu):
  """Creates an optimizer training op."""
  global_step = tf.train.get_or_create_global_step()

  learning_rate = tf.constant(value=init_lr, shape=[], dtype=tf.float32)

  # Implements linear decay of the learning rate.
  learning_rate = tf.train.polynomial_decay(
      learning_rate,
      global_step,
      num_train_steps,
      end_learning_rate=0.0,
      power=1.0,
      cycle=False)

  # Implements linear warmup. I.e., if global_step < num_warmup_steps, the
  # learning rate will be `global_step/num_warmup_steps * init_lr`.
  if num_warmup_steps:
    global_steps_int = tf.cast(global_step, tf.int32)
    warmup_steps_int = tf.constant(num_warmup_steps, dtype=tf.int32)

    global_steps_float = tf.cast(global_steps_int, tf.float32)
    warmup_steps_float = tf.cast(warmup_steps_int, tf.float32)

    warmup_percent_done = global_steps_float / warmup_steps_float
    warmup_learning_rate = init_lr * warmup_percent_done

    is_warmup = tf.cast(global_steps_int < warmup_steps_int, tf.float32)
    learning_rate = (
        (1.0 - is_warmup) * learning_rate + is_warmup * warmup_learning_rate)

  # It is recommended that you use this optimizer for fine tuning, since this
  # is how the model was trained (note that the Adam m/v variables are NOT
  # loaded from init_checkpoint.)
  optimizer = AdamWeightDecayOptimizer(
      learning_rate=learning_rate,
      weight_decay_rate=0.01,
      beta_1=0.9,
      beta_2=0.999,
      epsilon=1e-6,
      exclude_from_weight_decay=["LayerNorm", "layer_norm", "bias"])

  if use_tpu:
    optimizer = tf.contrib.tpu.CrossShardOptimizer(optimizer)

  tvars = tf.trainable_variables()
  grads = tf.gradients(loss, tvars)

  # This is how the model was pre-trained.
  (grads, _) = tf.clip_by_global_norm(grads, clip_norm=1.0)

  train_op = optimizer.apply_gradients(
      zip(grads, tvars), global_step=global_step)

  # Normally the global step update is done inside of `apply_gradients`.
  # However, `AdamWeightDecayOptimizer` doesn't do this. But if you use
  # a different optimizer, you should probably take this line out.
  new_global_step = global_step + 1
  train_op = tf.group(train_op, [global_step.assign(new_global_step)])
  return train_op


class AdamWeightDecayOptimizer(tf.train.Optimizer):
  """A basic Adam optimizer that includes "correct" L2 weight decay."""

  def __init__(self,
               learning_rate,
               weight_decay_rate=0.0,
               beta_1=0.9,
               beta_2=0.999,
               epsilon=1e-6,
               exclude_from_weight_decay=None,
               name="AdamWeightDecayOptimizer"):
    """Constructs a AdamWeightDecayOptimizer."""
    super(AdamWeightDecayOptimizer, self).__init__(False, name)

    self.learning_rate = learning_rate
    self.weight_decay_rate = weight_decay_rate
    self.beta_1 = beta_1
    self.beta_2 = beta_2
    self.epsilon = epsilon
    self.exclude_from_weight_decay = exclude_from_weight_decay

  def apply_gradients(self, grads_and_vars, global_step=None, name=None):
    """See base class."""
    assignments = []
    for (grad, param) in grads_and_vars:
      if grad is None or param is None:
        continue

      param_name = self._get_variable_name(param.name)

      m = tf.get_variable(
          name=param_name + "/adam_m",
          shape=param.shape.as_list(),
          dtype=tf.float32,
          trainable=False,
          initializer=tf.zeros_initializer())
      v = tf.get_variable(
          name=param_name + "/adam_v",
          shape=param.shape.as_list(),
          dtype=tf.float32,
          trainable=False,
          initializer=tf.zeros_initializer())

      # Standard Adam update.
      next_m = (
          tf.multiply(self.beta_1, m) + tf.multiply(1.0 - self.beta_1, grad))
      next_v = (
          tf.multiply(self.beta_2, v) + tf.multiply(1.0 - self.beta_2,
                                                    tf.square(grad)))

      update = next_m / (tf.sqrt(next_v) + self.epsilon)

      # Just adding the square of the weights to the loss function is *not*
      # the correct way of using L2 regularization/weight decay with Adam,
      # since that will interact with the m and v parameters in strange ways.
      #
      # Instead we want ot decay the weights in a manner that doesn't interact
      # with the m/v parameters. This is equivalent to adding the square
      # of the weights to the loss with plain (non-momentum) SGD.
      if self._do_use_weight_decay(param_name):
        update += self.weight_decay_rate * param

      update_with_lr = self.learning_rate * update

      next_param = param - update_with_lr

      assignments.extend(
          [param.assign(next_param),
           m.assign(next_m),
           v.assign(next_v)])
    return tf.group(*assignments, name=name)

  def _do_use_weight_decay(self, param_name):
    """Whether to use L2 weight decay for `param_name`."""
    if not self.weight_decay_rate:
      return False
    if self.exclude_from_weight_decay:
      for r in self.exclude_from_weight_decay:
        if re.search(r, param_name) is not None:
          return False
    return True

  def _get_variable_name(self, param_name):
    """Get the variable name from the tensor name."""
    m = re.match("^(.*):\\d+$", param_name)
    if m is not None:
      param_name = m.group(1)
    return param_name



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The main BERT model and related functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import copy
import json
import math
import re
import numpy as np
import six
import tensorflow as tf


class BertConfig(object):
  """Configuration for `BertModel`."""

  def __init__(self,
               vocab_size,
               hidden_size=768,
               num_hidden_layers=12,
               num_attention_heads=12,
               intermediate_size=3072,
               hidden_act="gelu",
               hidden_dropout_prob=0.1,
               attention_probs_dropout_prob=0.1,
               max_position_embeddings=512,
               type_vocab_size=16,
               initializer_range=0.02):
    """Constructs BertConfig.

    Args:
      vocab_size: Vocabulary size of `inputs_ids` in `BertModel`.
      hidden_size: Size of the encoder layers and the pooler layer.
      num_hidden_layers: Number of hidden layers in the Transformer encoder.
      num_attention_heads: Number of attention heads for each attention layer in
        the Transformer encoder.
      intermediate_size: The size of the "intermediate" (i.e., feed-forward)
        layer in the Transformer encoder.
      hidden_act: The non-linear activation function (function or string) in the
        encoder and pooler.
      hidden_dropout_prob: The dropout probability for all fully connected
        layers in the embeddings, encoder, and pooler.
      attention_probs_dropout_prob: The dropout ratio for the attention
        probabilities.
      max_position_embeddings: The maximum sequence length that this model might
        ever be used with. Typically set this to something large just in case
        (e.g., 512 or 1024 or 2048).
      type_vocab_size: The vocabulary size of the `token_type_ids` passed into
        `BertModel`.
      initializer_range: The stdev of the truncated_normal_initializer for
        initializing all weight matrices.
    """
    self.vocab_size = vocab_size
    self.hidden_size = hidden_size
    self.num_hidden_layers = num_hidden_layers
    self.num_attention_heads = num_attention_heads
    self.hidden_act = hidden_act
    self.intermediate_size = intermediate_size
    self.hidden_dropout_prob = hidden_dropout_prob
    self.attention_probs_dropout_prob = attention_probs_dropout_prob
    self.max_position_embeddings = max_position_embeddings
    self.type_vocab_size = type_vocab_size
    self.initializer_range = initializer_range

  @classmethod
  def from_dict(cls, json_object):
    """Constructs a `BertConfig` from a Python dictionary of parameters."""
    config = BertConfig(vocab_size=None)
    for (key, value) in six.iteritems(json_object):
      config.__dict__[key] = value
    return config

  @classmethod
  def from_json_file(cls, json_file):
    """Constructs a `BertConfig` from a json file of parameters."""
    with tf.gfile.GFile(json_file, "r") as reader:
      text = reader.read()
    return cls.from_dict(json.loads(text))

  def to_dict(self):
    """Serializes this instance to a Python dictionary."""
    output = copy.deepcopy(self.__dict__)
    return output

  def to_json_string(self):
    """Serializes this instance to a JSON string."""
    return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


class BertModel(object):
  """BERT model ("Bidirectional Encoder Representations from Transformers").

  Example usage:

  ```python
  # Already been converted into WordPiece token ids
  input_ids = tf.constant([[31, 51, 99], [15, 5, 0]])
  input_mask = tf.constant([[1, 1, 1], [1, 1, 0]])
  token_type_ids = tf.constant([[0, 0, 1], [0, 2, 0]])

  config = modeling.BertConfig(vocab_size=32000, hidden_size=512,
    num_hidden_layers=8, num_attention_heads=6, intermediate_size=1024)

  model = modeling.BertModel(config=config, is_training=True,
    input_ids=input_ids, input_mask=input_mask, token_type_ids=token_type_ids)

  label_embeddings = tf.get_variable(...)
  pooled_output = model.get_pooled_output()
  logits = tf.matmul(pooled_output, label_embeddings)
  ...
  ```
  """

  def __init__(self,
               config,
               is_training,
               input_ids,
               input_mask=None,
               token_type_ids=None,
               use_one_hot_embeddings=False,
               scope=None):
    """Constructor for BertModel.

    Args:
      config: `BertConfig` instance.
      is_training: bool. true for training model, false for eval model. Controls
        whether dropout will be applied.
      input_ids: int32 Tensor of shape [batch_size, seq_length].
      input_mask: (optional) int32 Tensor of shape [batch_size, seq_length].
      token_type_ids: (optional) int32 Tensor of shape [batch_size, seq_length].
      use_one_hot_embeddings: (optional) bool. Whether to use one-hot word
        embeddings or tf.embedding_lookup() for the word embeddings.
      scope: (optional) variable scope. Defaults to "bert".

    Raises:
      ValueError: The config is invalid or one of the input tensor shapes
        is invalid.
    """
    config = copy.deepcopy(config)
    if not is_training:
      config.hidden_dropout_prob = 0.0
      config.attention_probs_dropout_prob = 0.0

    input_shape = get_shape_list(input_ids, expected_rank=2)
    batch_size = input_shape[0]
    seq_length = input_shape[1]

    if input_mask is None:
      input_mask = tf.ones(shape=[batch_size, seq_length], dtype=tf.int32)

    if token_type_ids is None:
      token_type_ids = tf.zeros(shape=[batch_size, seq_length], dtype=tf.int32)

    with tf.variable_scope(scope, default_name="bert"):
      with tf.variable_scope("embeddings"):
        # Perform embedding lookup on the word ids.
        (self.embedding_output, self.embedding_table) = embedding_lookup(
            input_ids=input_ids,
            vocab_size=config.vocab_size,
            embedding_size=config.hidden_size,
            initializer_range=config.initializer_range,
            word_embedding_name="word_embeddings",
            use_one_hot_embeddings=use_one_hot_embeddings)

        # Add positional embeddings and token type embeddings, then layer
        # normalize and perform dropout.
        self.embedding_output = embedding_postprocessor(
            input_tensor=self.embedding_output,
            use_token_type=True,
            token_type_ids=token_type_ids,
            token_type_vocab_size=config.type_vocab_size,
            token_type_embedding_name="token_type_embeddings",
            use_position_embeddings=True,
            position_embedding_name="position_embeddings",
            initializer_range=config.initializer_range,
            max_position_embeddings=config.max_position_embeddings,
            dropout_prob=config.hidden_dropout_prob)

      with tf.variable_scope("encoder"):
        # This converts a 2D mask of shape [batch_size, seq_length] to a 3D
        # mask of shape [batch_size, seq_length, seq_length] which is used
        # for the attention scores.
        attention_mask = create_attention_mask_from_input_mask(
            input_ids, input_mask)

        # Run the stacked transformer.
        # `sequence_output` shape = [batch_size, seq_length, hidden_size].
        self.all_encoder_layers = transformer_model(
            input_tensor=self.embedding_output,
            attention_mask=attention_mask,
            hidden_size=config.hidden_size,
            num_hidden_layers=config.num_hidden_layers,
            num_attention_heads=config.num_attention_heads,
            intermediate_size=config.intermediate_size,
            intermediate_act_fn=get_activation(config.hidden_act),
            hidden_dropout_prob=config.hidden_dropout_prob,
            attention_probs_dropout_prob=config.attention_probs_dropout_prob,
            initializer_range=config.initializer_range,
            do_return_all_layers=True)

      self.sequence_output = self.all_encoder_layers[-1]
      # The "pooler" converts the encoded sequence tensor of shape
      # [batch_size, seq_length, hidden_size] to a tensor of shape
      # [batch_size, hidden_size]. This is necessary for segment-level
      # (or segment-pair-level) classification tasks where we need a fixed
      # dimensional representation of the segment.
      with tf.variable_scope("pooler"):
        # We "pool" the model by simply taking the hidden state corresponding
        # to the first token. We assume that this has been pre-trained
        first_token_tensor = tf.squeeze(self.sequence_output[:, 0:1, :], axis=1)
        self.pooled_output = tf.layers.dense(
            first_token_tensor,
            config.hidden_size,
            activation=tf.tanh,
            kernel_initializer=create_initializer(config.initializer_range))

  def get_pooled_output(self):
    return self.pooled_output

  def get_sequence_output(self):
    """Gets final hidden layer of encoder.

    Returns:
      float Tensor of shape [batch_size, seq_length, hidden_size] corresponding
      to the final hidden of the transformer encoder.
    """
    return self.sequence_output

  def get_all_encoder_layers(self):
    return self.all_encoder_layers

  def get_embedding_output(self):
    """Gets output of the embedding lookup (i.e., input to the transformer).

    Returns:
      float Tensor of shape [batch_size, seq_length, hidden_size] corresponding
      to the output of the embedding layer, after summing the word
      embeddings with the positional embeddings and the token type embeddings,
      then performing layer normalization. This is the input to the transformer.
    """
    return self.embedding_output

  def get_embedding_table(self):
    return self.embedding_table


def gelu(x):
  """Gaussian Error Linear Unit.

  This is a smoother version of the RELU.
  Original paper: https://arxiv.org/abs/1606.08415
  Args:
    x: float Tensor to perform activation.

  Returns:
    `x` with the GELU activation applied.
  """
  cdf = 0.5 * (1.0 + tf.tanh(
      (np.sqrt(2 / np.pi) * (x + 0.044715 * tf.pow(x, 3)))))
  return x * cdf


def get_activation(activation_string):
  """Maps a string to a Python function, e.g., "relu" => `tf.nn.relu`.

  Args:
    activation_string: String name of the activation function.

  Returns:
    A Python function corresponding to the activation function. If
    `activation_string` is None, empty, or "linear", this will return None.
    If `activation_string` is not a string, it will return `activation_string`.

  Raises:
    ValueError: The `activation_string` does not correspond to a known
      activation.
  """

  # We assume that anything that"s not a string is already an activation
  # function, so we just return it.
  if not isinstance(activation_string, six.string_types):
    return activation_string

  if not activation_string:
    return None

  act = activation_string.lower()
  if act == "linear":
    return None
  elif act == "relu":
    return tf.nn.relu
  elif act == "gelu":
    return gelu
  elif act == "tanh":
    return tf.tanh
  else:
    raise ValueError("Unsupported activation: %s" % act)


def get_assignment_map_from_checkpoint(tvars, init_checkpoint):
  """Compute the union of the current variables and checkpoint variables."""
  assignment_map = {}
  initialized_variable_names = {}

  name_to_variable = collections.OrderedDict()
  for var in tvars:
    name = var.name
    m = re.match("^(.*):\\d+$", name)
    if m is not None:
      name = m.group(1)
    name_to_variable[name] = var

  init_vars = tf.train.list_variables(init_checkpoint)

  assignment_map = collections.OrderedDict()
  for x in init_vars:
    (name, var) = (x[0], x[1])
    if name not in name_to_variable:
      continue
    assignment_map[name] = name
    initialized_variable_names[name] = 1
    initialized_variable_names[name + ":0"] = 1

  return (assignment_map, initialized_variable_names)


def dropout(input_tensor, dropout_prob):
  """Perform dropout.

  Args:
    input_tensor: float Tensor.
    dropout_prob: Python float. The probability of dropping out a value (NOT of
      *keeping* a dimension as in `tf.nn.dropout`).

  Returns:
    A version of `input_tensor` with dropout applied.
  """
  if dropout_prob is None or dropout_prob == 0.0:
    return input_tensor

  output = tf.nn.dropout(input_tensor, 1.0 - dropout_prob)
  return output


def layer_norm(input_tensor, name=None):
  """Run layer normalization on the last dimension of the tensor."""
  return tf.contrib.layers.layer_norm(
      inputs=input_tensor, begin_norm_axis=-1, begin_params_axis=-1, scope=name)


def layer_norm_and_dropout(input_tensor, dropout_prob, name=None):
  """Runs layer normalization followed by dropout."""
  output_tensor = layer_norm(input_tensor, name)
  output_tensor = dropout(output_tensor, dropout_prob)
  return output_tensor


def create_initializer(initializer_range=0.02):
  """Creates a `truncated_normal_initializer` with the given range."""
  return tf.truncated_normal_initializer(stddev=initializer_range)


def embedding_lookup(input_ids,
                     vocab_size,
                     embedding_size=128,
                     initializer_range=0.02,
                     word_embedding_name="word_embeddings",
                     use_one_hot_embeddings=False):
  """Looks up words embeddings for id tensor.

  Args:
    input_ids: int32 Tensor of shape [batch_size, seq_length] containing word
      ids.
    vocab_size: int. Size of the embedding vocabulary.
    embedding_size: int. Width of the word embeddings.
    initializer_range: float. Embedding initialization range.
    word_embedding_name: string. Name of the embedding table.
    use_one_hot_embeddings: bool. If True, use one-hot method for word
      embeddings. If False, use `tf.gather()`.

  Returns:
    float Tensor of shape [batch_size, seq_length, embedding_size].
  """
  # This function assumes that the input is of shape [batch_size, seq_length,
  # num_inputs].
  #
  # If the input is a 2D tensor of shape [batch_size, seq_length], we
  # reshape to [batch_size, seq_length, 1].
  if input_ids.shape.ndims == 2:
    input_ids = tf.expand_dims(input_ids, axis=[-1])

  embedding_table = tf.get_variable(
      name=word_embedding_name,
      shape=[vocab_size, embedding_size],
      initializer=create_initializer(initializer_range))

  flat_input_ids = tf.reshape(input_ids, [-1])
  if use_one_hot_embeddings:
    one_hot_input_ids = tf.one_hot(flat_input_ids, depth=vocab_size)
    output = tf.matmul(one_hot_input_ids, embedding_table)
  else:
    output = tf.gather(embedding_table, flat_input_ids)

  input_shape = get_shape_list(input_ids)

  output = tf.reshape(output,
                      input_shape[0:-1] + [input_shape[-1] * embedding_size])
  return (output, embedding_table)


def embedding_postprocessor(input_tensor,
                            use_token_type=False,
                            token_type_ids=None,
                            token_type_vocab_size=16,
                            token_type_embedding_name="token_type_embeddings",
                            use_position_embeddings=True,
                            position_embedding_name="position_embeddings",
                            initializer_range=0.02,
                            max_position_embeddings=512,
                            dropout_prob=0.1):
  """Performs various post-processing on a word embedding tensor.

  Args:
    input_tensor: float Tensor of shape [batch_size, seq_length,
      embedding_size].
    use_token_type: bool. Whether to add embeddings for `token_type_ids`.
    token_type_ids: (optional) int32 Tensor of shape [batch_size, seq_length].
      Must be specified if `use_token_type` is True.
    token_type_vocab_size: int. The vocabulary size of `token_type_ids`.
    token_type_embedding_name: string. The name of the embedding table variable
      for token type ids.
    use_position_embeddings: bool. Whether to add position embeddings for the
      position of each token in the sequence.
    position_embedding_name: string. The name of the embedding table variable
      for positional embeddings.
    initializer_range: float. Range of the weight initialization.
    max_position_embeddings: int. Maximum sequence length that might ever be
      used with this model. This can be longer than the sequence length of
      input_tensor, but cannot be shorter.
    dropout_prob: float. Dropout probability applied to the final output tensor.

  Returns:
    float tensor with same shape as `input_tensor`.

  Raises:
    ValueError: One of the tensor shapes or input values is invalid.
  """
  input_shape = get_shape_list(input_tensor, expected_rank=3)
  batch_size = input_shape[0]
  seq_length = input_shape[1]
  width = input_shape[2]

  output = input_tensor

  if use_token_type:
    if token_type_ids is None:
      raise ValueError("`token_type_ids` must be specified if"
                       "`use_token_type` is True.")
    token_type_table = tf.get_variable(
        name=token_type_embedding_name,
        shape=[token_type_vocab_size, width],
        initializer=create_initializer(initializer_range))
    # This vocab will be small so we always do one-hot here, since it is always
    # faster for a small vocabulary.
    flat_token_type_ids = tf.reshape(token_type_ids, [-1])
    one_hot_ids = tf.one_hot(flat_token_type_ids, depth=token_type_vocab_size)
    token_type_embeddings = tf.matmul(one_hot_ids, token_type_table)
    token_type_embeddings = tf.reshape(token_type_embeddings,
                                       [batch_size, seq_length, width])
    output += token_type_embeddings

  if use_position_embeddings:
    assert_op = tf.assert_less_equal(seq_length, max_position_embeddings)
    with tf.control_dependencies([assert_op]):
      full_position_embeddings = tf.get_variable(
          name=position_embedding_name,
          shape=[max_position_embeddings, width],
          initializer=create_initializer(initializer_range))
      # Since the position embedding table is a learned variable, we create it
      # using a (long) sequence length `max_position_embeddings`. The actual
      # sequence length might be shorter than this, for faster training of
      # tasks that do not have long sequences.
      #
      # So `full_position_embeddings` is effectively an embedding table
      # for position [0, 1, 2, ..., max_position_embeddings-1], and the current
      # sequence has positions [0, 1, 2, ... seq_length-1], so we can just
      # perform a slice.
      position_embeddings = tf.slice(full_position_embeddings, [0, 0],
                                     [seq_length, -1])
      num_dims = len(output.shape.as_list())

      # Only the last two dimensions are relevant (`seq_length` and `width`), so
      # we broadcast among the first dimensions, which is typically just
      # the batch size.
      position_broadcast_shape = []
      for _ in range(num_dims - 2):
        position_broadcast_shape.append(1)
      position_broadcast_shape.extend([seq_length, width])
      position_embeddings = tf.reshape(position_embeddings,
                                       position_broadcast_shape)
      output += position_embeddings

  output = layer_norm_and_dropout(output, dropout_prob)
  return output


def create_attention_mask_from_input_mask(from_tensor, to_mask):
  """Create 3D attention mask from a 2D tensor mask.

  Args:
    from_tensor: 2D or 3D Tensor of shape [batch_size, from_seq_length, ...].
    to_mask: int32 Tensor of shape [batch_size, to_seq_length].

  Returns:
    float Tensor of shape [batch_size, from_seq_length, to_seq_length].
  """
  from_shape = get_shape_list(from_tensor, expected_rank=[2, 3])
  batch_size = from_shape[0]
  from_seq_length = from_shape[1]

  to_shape = get_shape_list(to_mask, expected_rank=2)
  to_seq_length = to_shape[1]

  to_mask = tf.cast(
      tf.reshape(to_mask, [batch_size, 1, to_seq_length]), tf.float32)

  # We don't assume that `from_tensor` is a mask (although it could be). We
  # don't actually care if we attend *from* padding tokens (only *to* padding)
  # tokens so we create a tensor of all ones.
  #
  # `broadcast_ones` = [batch_size, from_seq_length, 1]
  broadcast_ones = tf.ones(
      shape=[batch_size, from_seq_length, 1], dtype=tf.float32)

  # Here we broadcast along two dimensions to create the mask.
  mask = broadcast_ones * to_mask

  return mask


def attention_layer(from_tensor,
                    to_tensor,
                    attention_mask=None,
                    num_attention_heads=1,
                    size_per_head=512,
                    query_act=None,
                    key_act=None,
                    value_act=None,
                    attention_probs_dropout_prob=0.0,
                    initializer_range=0.02,
                    do_return_2d_tensor=False,
                    batch_size=None,
                    from_seq_length=None,
                    to_seq_length=None):
  """Performs multi-headed attention from `from_tensor` to `to_tensor`.

  This is an implementation of multi-headed attention based on "Attention
  is all you Need". If `from_tensor` and `to_tensor` are the same, then
  this is self-attention. Each timestep in `from_tensor` attends to the
  corresponding sequence in `to_tensor`, and returns a fixed-with vector.

  This function first projects `from_tensor` into a "query" tensor and
  `to_tensor` into "key" and "value" tensors. These are (effectively) a list
  of tensors of length `num_attention_heads`, where each tensor is of shape
  [batch_size, seq_length, size_per_head].

  Then, the query and key tensors are dot-producted and scaled. These are
  softmaxed to obtain attention probabilities. The value tensors are then
  interpolated by these probabilities, then concatenated back to a single
  tensor and returned.

  In practice, the multi-headed attention are done with transposes and
  reshapes rather than actual separate tensors.

  Args:
    from_tensor: float Tensor of shape [batch_size, from_seq_length,
      from_width].
    to_tensor: float Tensor of shape [batch_size, to_seq_length, to_width].
    attention_mask: (optional) int32 Tensor of shape [batch_size,
      from_seq_length, to_seq_length]. The values should be 1 or 0. The
      attention scores will effectively be set to -infinity for any positions in
      the mask that are 0, and will be unchanged for positions that are 1.
    num_attention_heads: int. Number of attention heads.
    size_per_head: int. Size of each attention head.
    query_act: (optional) Activation function for the query transform.
    key_act: (optional) Activation function for the key transform.
    value_act: (optional) Activation function for the value transform.
    attention_probs_dropout_prob: (optional) float. Dropout probability of the
      attention probabilities.
    initializer_range: float. Range of the weight initializer.
    do_return_2d_tensor: bool. If True, the output will be of shape [batch_size
      * from_seq_length, num_attention_heads * size_per_head]. If False, the
      output will be of shape [batch_size, from_seq_length, num_attention_heads
      * size_per_head].
    batch_size: (Optional) int. If the input is 2D, this might be the batch size
      of the 3D version of the `from_tensor` and `to_tensor`.
    from_seq_length: (Optional) If the input is 2D, this might be the seq length
      of the 3D version of the `from_tensor`.
    to_seq_length: (Optional) If the input is 2D, this might be the seq length
      of the 3D version of the `to_tensor`.

  Returns:
    float Tensor of shape [batch_size, from_seq_length,
      num_attention_heads * size_per_head]. (If `do_return_2d_tensor` is
      true, this will be of shape [batch_size * from_seq_length,
      num_attention_heads * size_per_head]).

  Raises:
    ValueError: Any of the arguments or tensor shapes are invalid.
  """

  def transpose_for_scores(input_tensor, batch_size, num_attention_heads,
                           seq_length, width):
    output_tensor = tf.reshape(
        input_tensor, [batch_size, seq_length, num_attention_heads, width])

    output_tensor = tf.transpose(output_tensor, [0, 2, 1, 3])
    return output_tensor

  from_shape = get_shape_list(from_tensor, expected_rank=[2, 3])
  to_shape = get_shape_list(to_tensor, expected_rank=[2, 3])

  if len(from_shape) != len(to_shape):
    raise ValueError(
        "The rank of `from_tensor` must match the rank of `to_tensor`.")

  if len(from_shape) == 3:
    batch_size = from_shape[0]
    from_seq_length = from_shape[1]
    to_seq_length = to_shape[1]
  elif len(from_shape) == 2:
    if (batch_size is None or from_seq_length is None or to_seq_length is None):
      raise ValueError(
          "When passing in rank 2 tensors to attention_layer, the values "
          "for `batch_size`, `from_seq_length`, and `to_seq_length` "
          "must all be specified.")

  # Scalar dimensions referenced here:
  #   B = batch size (number of sequences)
  #   F = `from_tensor` sequence length
  #   T = `to_tensor` sequence length
  #   N = `num_attention_heads`
  #   H = `size_per_head`

  from_tensor_2d = reshape_to_matrix(from_tensor)
  to_tensor_2d = reshape_to_matrix(to_tensor)

  # `query_layer` = [B*F, N*H]
  query_layer = tf.layers.dense(
      from_tensor_2d,
      num_attention_heads * size_per_head,
      activation=query_act,
      name="query",
      kernel_initializer=create_initializer(initializer_range))

  # `key_layer` = [B*T, N*H]
  key_layer = tf.layers.dense(
      to_tensor_2d,
      num_attention_heads * size_per_head,
      activation=key_act,
      name="key",
      kernel_initializer=create_initializer(initializer_range))

  # `value_layer` = [B*T, N*H]
  value_layer = tf.layers.dense(
      to_tensor_2d,
      num_attention_heads * size_per_head,
      activation=value_act,
      name="value",
      kernel_initializer=create_initializer(initializer_range))

  # `query_layer` = [B, N, F, H]
  query_layer = transpose_for_scores(query_layer, batch_size,
                                     num_attention_heads, from_seq_length,
                                     size_per_head)

  # `key_layer` = [B, N, T, H]
  key_layer = transpose_for_scores(key_layer, batch_size, num_attention_heads,
                                   to_seq_length, size_per_head)

  # Take the dot product between "query" and "key" to get the raw
  # attention scores.
  # `attention_scores` = [B, N, F, T]
  attention_scores = tf.matmul(query_layer, key_layer, transpose_b=True)
  attention_scores = tf.multiply(attention_scores,
                                 1.0 / math.sqrt(float(size_per_head)))

  if attention_mask is not None:
    # `attention_mask` = [B, 1, F, T]
    attention_mask = tf.expand_dims(attention_mask, axis=[1])

    # Since attention_mask is 1.0 for positions we want to attend and 0.0 for
    # masked positions, this operation will create a tensor which is 0.0 for
    # positions we want to attend and -10000.0 for masked positions.
    adder = (1.0 - tf.cast(attention_mask, tf.float32)) * -10000.0

    # Since we are adding it to the raw scores before the softmax, this is
    # effectively the same as removing these entirely.
    attention_scores += adder

  # Normalize the attention scores to probabilities.
  # `attention_probs` = [B, N, F, T]
  attention_probs = tf.nn.softmax(attention_scores)

  # This is actually dropping out entire tokens to attend to, which might
  # seem a bit unusual, but is taken from the original Transformer paper.
  attention_probs = dropout(attention_probs, attention_probs_dropout_prob)

  # `value_layer` = [B, T, N, H]
  value_layer = tf.reshape(
      value_layer,
      [batch_size, to_seq_length, num_attention_heads, size_per_head])

  # `value_layer` = [B, N, T, H]
  value_layer = tf.transpose(value_layer, [0, 2, 1, 3])

  # `context_layer` = [B, N, F, H]
  context_layer = tf.matmul(attention_probs, value_layer)

  # `context_layer` = [B, F, N, H]
  context_layer = tf.transpose(context_layer, [0, 2, 1, 3])

  if do_return_2d_tensor:
    # `context_layer` = [B*F, N*H]
    context_layer = tf.reshape(
        context_layer,
        [batch_size * from_seq_length, num_attention_heads * size_per_head])
  else:
    # `context_layer` = [B, F, N*H]
    context_layer = tf.reshape(
        context_layer,
        [batch_size, from_seq_length, num_attention_heads * size_per_head])

  return context_layer


def transformer_model(input_tensor,
                      attention_mask=None,
                      hidden_size=768,
                      num_hidden_layers=12,
                      num_attention_heads=12,
                      intermediate_size=3072,
                      intermediate_act_fn=gelu,
                      hidden_dropout_prob=0.1,
                      attention_probs_dropout_prob=0.1,
                      initializer_range=0.02,
                      do_return_all_layers=False):
  """Multi-headed, multi-layer Transformer from "Attention is All You Need".

  This is almost an exact implementation of the original Transformer encoder.

  See the original paper:
  https://arxiv.org/abs/1706.03762

  Also see:
  https://github.com/tensorflow/tensor2tensor/blob/master/tensor2tensor/models/transformer.py

  Args:
    input_tensor: float Tensor of shape [batch_size, seq_length, hidden_size].
    attention_mask: (optional) int32 Tensor of shape [batch_size, seq_length,
      seq_length], with 1 for positions that can be attended to and 0 in
      positions that should not be.
    hidden_size: int. Hidden size of the Transformer.
    num_hidden_layers: int. Number of layers (blocks) in the Transformer.
    num_attention_heads: int. Number of attention heads in the Transformer.
    intermediate_size: int. The size of the "intermediate" (a.k.a., feed
      forward) layer.
    intermediate_act_fn: function. The non-linear activation function to apply
      to the output of the intermediate/feed-forward layer.
    hidden_dropout_prob: float. Dropout probability for the hidden layers.
    attention_probs_dropout_prob: float. Dropout probability of the attention
      probabilities.
    initializer_range: float. Range of the initializer (stddev of truncated
      normal).
    do_return_all_layers: Whether to also return all layers or just the final
      layer.

  Returns:
    float Tensor of shape [batch_size, seq_length, hidden_size], the final
    hidden layer of the Transformer.

  Raises:
    ValueError: A Tensor shape or parameter is invalid.
  """
  if hidden_size % num_attention_heads != 0:
    raise ValueError(
        "The hidden size (%d) is not a multiple of the number of attention "
        "heads (%d)" % (hidden_size, num_attention_heads))

  attention_head_size = int(hidden_size / num_attention_heads)
  input_shape = get_shape_list(input_tensor, expected_rank=3)
  batch_size = input_shape[0]
  seq_length = input_shape[1]
  input_width = input_shape[2]

  # The Transformer performs sum residuals on all layers so the input needs
  # to be the same as the hidden size.
  if input_width != hidden_size:
    raise ValueError("The width of the input tensor (%d) != hidden size (%d)" %
                     (input_width, hidden_size))

  # We keep the representation as a 2D tensor to avoid re-shaping it back and
  # forth from a 3D tensor to a 2D tensor. Re-shapes are normally free on
  # the GPU/CPU but may not be free on the TPU, so we want to minimize them to
  # help the optimizer.
  prev_output = reshape_to_matrix(input_tensor)

  all_layer_outputs = []
  for layer_idx in range(num_hidden_layers):
    with tf.variable_scope("layer_%d" % layer_idx):
      layer_input = prev_output

      with tf.variable_scope("attention"):
        attention_heads = []
        with tf.variable_scope("self"):
          attention_head = attention_layer(
              from_tensor=layer_input,
              to_tensor=layer_input,
              attention_mask=attention_mask,
              num_attention_heads=num_attention_heads,
              size_per_head=attention_head_size,
              attention_probs_dropout_prob=attention_probs_dropout_prob,
              initializer_range=initializer_range,
              do_return_2d_tensor=True,
              batch_size=batch_size,
              from_seq_length=seq_length,
              to_seq_length=seq_length)
          attention_heads.append(attention_head)

        attention_output = None
        if len(attention_heads) == 1:
          attention_output = attention_heads[0]
        else:
          # In the case where we have other sequences, we just concatenate
          # them to the self-attention head before the projection.
          attention_output = tf.concat(attention_heads, axis=-1)

        # Run a linear projection of `hidden_size` then add a residual
        # with `layer_input`.
        with tf.variable_scope("output"):
          attention_output = tf.layers.dense(
              attention_output,
              hidden_size,
              kernel_initializer=create_initializer(initializer_range))
          attention_output = dropout(attention_output, hidden_dropout_prob)
          attention_output = layer_norm(attention_output + layer_input)

      # The activation is only applied to the "intermediate" hidden layer.
      with tf.variable_scope("intermediate"):
        intermediate_output = tf.layers.dense(
            attention_output,
            intermediate_size,
            activation=intermediate_act_fn,
            kernel_initializer=create_initializer(initializer_range))

      # Down-project back to `hidden_size` then add the residual.
      with tf.variable_scope("output"):
        layer_output = tf.layers.dense(
            intermediate_output,
            hidden_size,
            kernel_initializer=create_initializer(initializer_range))
        layer_output = dropout(layer_output, hidden_dropout_prob)
        layer_output = layer_norm(layer_output + attention_output)
        prev_output = layer_output
        all_layer_outputs.append(layer_output)

  if do_return_all_layers:
    final_outputs = []
    for layer_output in all_layer_outputs:
      final_output = reshape_from_matrix(layer_output, input_shape)
      final_outputs.append(final_output)
    return final_outputs
  else:
    final_output = reshape_from_matrix(prev_output, input_shape)
    return final_output


def get_shape_list(tensor, expected_rank=None, name=None):
  """Returns a list of the shape of tensor, preferring static dimensions.

  Args:
    tensor: A tf.Tensor object to find the shape of.
    expected_rank: (optional) int. The expected rank of `tensor`. If this is
      specified and the `tensor` has a different rank, and exception will be
      thrown.
    name: Optional name of the tensor for the error message.

  Returns:
    A list of dimensions of the shape of tensor. All static dimensions will
    be returned as python integers, and dynamic dimensions will be returned
    as tf.Tensor scalars.
  """
  if name is None:
    name = tensor.name

  if expected_rank is not None:
    assert_rank(tensor, expected_rank, name)

  shape = tensor.shape.as_list()

  non_static_indexes = []
  for (index, dim) in enumerate(shape):
    if dim is None:
      non_static_indexes.append(index)

  if not non_static_indexes:
    return shape

  dyn_shape = tf.shape(tensor)
  for index in non_static_indexes:
    shape[index] = dyn_shape[index]
  return shape


def reshape_to_matrix(input_tensor):
  """Reshapes a >= rank 2 tensor to a rank 2 tensor (i.e., a matrix)."""
  ndims = input_tensor.shape.ndims
  if ndims < 2:
    raise ValueError("Input tensor must have at least rank 2. Shape = %s" %
                     (input_tensor.shape))
  if ndims == 2:
    return input_tensor

  width = input_tensor.shape[-1]
  output_tensor = tf.reshape(input_tensor, [-1, width])
  return output_tensor


def reshape_from_matrix(output_tensor, orig_shape_list):
  """Reshapes a rank 2 tensor back to its original rank >= 2 tensor."""
  if len(orig_shape_list) == 2:
    return output_tensor

  output_shape = get_shape_list(output_tensor)

  orig_dims = orig_shape_list[0:-1]
  width = output_shape[-1]

  return tf.reshape(output_tensor, orig_dims + [width])


def assert_rank(tensor, expected_rank, name=None):
  """Raises an exception if the tensor rank is not of the expected rank.

  Args:
    tensor: A tf.Tensor to check the rank of.
    expected_rank: Python integer or list of integers, expected rank.
    name: Optional name of the tensor for the error message.

  Raises:
    ValueError: If the expected shape doesn't match the actual shape.
  """
  if name is None:
    name = tensor.name

  expected_rank_dict = {}
  if isinstance(expected_rank, six.integer_types):
    expected_rank_dict[expected_rank] = True
  else:
    for x in expected_rank:
      expected_rank_dict[x] = True

  actual_rank = tensor.shape.ndims
  if actual_rank not in expected_rank_dict:
    scope_name = tf.get_variable_scope().name
    raise ValueError(
        "For the tensor `%s` in scope `%s`, the actual rank "
        "`%d` (shape = %s) is not equal to the expected rank `%s`" %
        (name, scope_name, actual_rank, str(tensor.shape), str(expected_rank)))



import os
import tensorflow as tf
import numpy as np

curr_dir = os.path.dirname(os.path.realpath(__file__))

from qiznlp.common.modules.common_layers import mask_nonpad_from_embedding
from qiznlp.common.modules.embedding import embedding
from qiznlp.common.modules.birnn import Bi_RNN
from qiznlp.common.modules.idcnn import IDCNN
import qiznlp.common.utils as utils

conf = utils.dict2obj({
    'vocab_size': 1847,
    'embed_size': 300,
    'label_size': 30,
    'birnn_hidden_size': 300,
    'dropout_rate': 0.2,
    'lr': 1e-3,
    'pretrain_emb': None,
    # 'pretrain_emb': np.load(f'{curr_dir}/pretrain_word_emb300.npy'),
})


class Model(object):
    def __init__(self, build_graph=True, **kwargs):
        self.conf = conf
        self.run_model = kwargs.get('run_model', None)  # acquire outside run_model instance
        if build_graph:
            # build placeholder
            self.build_placeholder()
            # build model
            self.model_name = kwargs.get('model_name', 'birnn')
            {
                'birnn': self.build_model1,
                'idcnn': self.build_model2,
                'bert_crf': self.build_model3,
                # add new here
            }[self.model_name]()
            print(f'model_name: {self.model_name} build graph ok!')

    def build_placeholder(self):
        # placeholder
        # 原则上模型输入输出不变，不需换新model
        self.s1 = tf.placeholder(tf.int32, [None, None], name='s1')
        self.ner_label = tf.placeholder(tf.int32, [None, None], name='ner_label')
        self.dropout_rate = tf.placeholder(tf.float32, name='dropout_rate')

    def build_model1(self):
        # embedding
        # [batch,len,embed]
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, name='embedding', pretrain_embedding=conf.pretrain_emb)
        s1_mask = mask_nonpad_from_embedding(s1_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s1_seqlen = tf.cast(tf.reduce_sum(s1_mask, axis=-1), tf.int32)  # [batch]

        # encoder
        encoder_input = s1_embed
        encoder_input = tf.layers.dropout(encoder_input, rate=self.dropout_rate)  # dropout

        with tf.variable_scope('rnn_1'):
            self.bilstm_encoder1 = Bi_RNN(cell_name='LSTMCell', hidden_size=conf.birnn_hidden_size, dropout_rate=self.dropout_rate)
            encoder_output, _ = self.bilstm_encoder1(encoder_input, s1_seqlen)  # [batch,len,2hid]
        with tf.variable_scope('rnn_2'):
            self.bilstm_encoder2 = Bi_RNN(cell_name='LSTMCell', hidden_size=conf.birnn_hidden_size, dropout_rate=self.dropout_rate)
            encoder_output, _ = self.bilstm_encoder2(encoder_output, s1_seqlen)  # [batch,len,2hid]

        # logits
        cls_logits = tf.layers.dense(encoder_output, conf.label_size, activation=None, use_bias=True, name='cls_logits')  # [batch,length,label]

        # crf
        log_likelihood, transition_params = tf.contrib.crf.crf_log_likelihood(cls_logits, self.ner_label, tf.to_int32(s1_seqlen))
        crf_loss = tf.reduce_mean(-log_likelihood)

        # loss
        self.loss = crf_loss

        # crf decode
        ner_pred, ner_prob = tf.contrib.crf.crf_decode(cls_logits, transition_params, tf.to_int32(s1_seqlen))

        self.ner_prob = tf.identity(ner_prob, name='ner_prob')  # [batch]
        self.ner_pred = tf.identity(ner_pred, name='ner_pred')  # [batch,len]

        with tf.name_scope('accuracy'):
            ner_acc = tf.cast(tf.equal(self.ner_pred, self.ner_label), tf.float32) * s1_mask  # [batch,length]
            ner_acc = tf.reduce_sum(ner_acc, axis=-1)  # [batch]
            ner_acc = ner_acc / tf.cast(s1_seqlen, tf.float32)  # [batch]
            self.accuracy = tf.reduce_mean(ner_acc, axis=0)  # scalar

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model2(self):
        # embedding
        # [batch,len,embed]
        # pretrained_word_embeddings = np.load(f'{curr_dir}/pretrain_emb_300.npy')
        pretrained_word_embeddings = None
        s1_embed, _ = embedding(tf.expand_dims(self.s1, -1), conf.vocab_size, conf.embed_size, name='embedding', pretrain_embedding=pretrained_word_embeddings)
        s1_mask = mask_nonpad_from_embedding(s1_embed)  # [batch,len] 1 for nonpad; 0 for pad
        s1_seqlen = tf.cast(tf.reduce_sum(s1_mask, axis=-1), tf.int32)  # [batch]

        # encoder
        encoder_input = s1_embed
        encoder_input = tf.layers.dropout(encoder_input, self.dropout_rate)  # dropout

        idcnn_net = IDCNN()
        encoder_output = idcnn_net(encoder_input)
        encoder_output = tf.layers.dropout(encoder_output, self.dropout_rate)  # dropout

        # logits
        cls_logits = tf.layers.dense(encoder_output, conf.label_size, activation=None, use_bias=True, name='cls_logits')  # [batch,length,label]

        # crf
        log_likelihood, transition_params = tf.contrib.crf.crf_log_likelihood(cls_logits, self.ner_label, tf.to_int32(s1_seqlen))
        crf_loss = tf.reduce_mean(-log_likelihood)

        # loss
        self.loss = crf_loss

        # crf decode
        ner_pred, ner_prob = tf.contrib.crf.crf_decode(cls_logits, transition_params, tf.to_int32(s1_seqlen))

        self.ner_prob = tf.identity(ner_prob, name='ner_prob')  # [batch]
        self.ner_pred = tf.identity(ner_pred, name='ner_pred')  # [batch,len]

        with tf.name_scope('accuracy'):
            ner_acc = tf.cast(tf.equal(self.ner_pred, self.ner_label), tf.float32) * s1_mask  # [batch,len]
            ner_acc = tf.reduce_sum(ner_acc, axis=-1)  # [batch]
            ner_acc = ner_acc / tf.cast(s1_seqlen, tf.float32)  # [batch]
            self.accuracy = tf.reduce_mean(ner_acc, axis=0)  # scalar

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    def build_model3(self):
        from qiznlp.common.modules.bert_model import BERT, get_tokenizer
        # adapt for bert
        bert_model_dir = f'{curr_dir}/../common/modules/bert/chinese_L-12_H-768_A-12'
        self.s1_input_mask = tf.placeholder(tf.int32, [None, None], name='s1_input_mask')
        self.s1_segment_ids = tf.placeholder(tf.int32, [None, None], name='s1_segment_ids')
        self.conf.lr = 2e-5  # official recommend for bert fintune
        bert_max_length = 128
        self.run_model.token2id_dct['char2id'] = utils.Any2Id.from_file(os.path.join(bert_model_dir, 'vocab.txt'), use_line_no=True)  # Bert vocab dict
        char2id = self.run_model.token2id_dct['char2id']
        bmeo2id = self.run_model.token2id_dct['bmeo2id']
        char2id['<unk>'] = char2id.pop('[UNK]')
        char2id['<pad>'] = char2id.pop('[PAD]')

        bert_tokenizer = get_tokenizer(os.path.join(bert_model_dir, 'vocab.txt'))
        self.run_model.tokenize = bert_tokenizer.tokenize
        self.run_model.cut = lambda t: ' '.join(self.run_model.tokenize(t))

        # 根据bert输入重写实例的feed_dict函数
        def _create_feed_dict_from_features(self, features, mode = 'train'):
            s1 = features['s1'].tolist()
            ner_label = features['ner_label'].tolist()
            new_s1 = []
            new_ner_label = []
            s1_input_mask = []
            s1_segment_ids = []

            for s1_, ner_label_ in zip(s1, ner_label):  # 遍历batch
                # remove pad
                i = len(s1_)-1
                while s1_[i] == char2id['<pad>']:  # [PAD]
                    i -= 1
                s1_ = s1_[:i+1]
                ner_label_ = ner_label_[:i+1]

                s1_ = [char2id['[CLS]']] + s1_[:bert_max_length-2] + [char2id['[SEP]']]  # [CLS] + [] + [SEP]
                ner_label_ = [bmeo2id['O']] + ner_label_[:bert_max_length-2] + [bmeo2id['O']]  # [O] + [] + [O]
                input_mask_ = [1] * len(s1_)
                segment_ids_ = [0] * len(s1_)

                new_s1.append(s1_)
                new_ner_label.append(ner_label_)
                s1_input_mask.append(input_mask_)
                s1_segment_ids.append(segment_ids_)

            feed_dict = {
                self.s1: utils.pad_sequences(new_s1, padding='post'),
                self.s1_input_mask: utils.pad_sequences(s1_input_mask, padding='post'),
                self.s1_segment_ids: utils.pad_sequences(s1_segment_ids, padding='post'),
                self.ner_label: utils.pad_sequences(new_ner_label, padding='post'),
            }
            if mode == 'train': feed_dict['num'] = len(features['s1'])
            feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
            return feed_dict

        def _create_feed_dict_from_raw(self, batch_s1, batch_ner_label, token2id_dct, mode='infer'):
            char2id = token2id_dct['char2id']
            bmeo2id = token2id_dct['bmeo2id']

            feed_s1 = [self.sent2ids(s1, char2id) for s1 in batch_s1]
            feed_input_mask = []
            feed_segment_ids = []
            for i, s1_ in enumerate(feed_s1):
                feed_s1[i] = [char2id['[CLS]']] + s1_[:bert_max_length - 2] + [char2id['[SEP]']]  # [CLS] + [] + [SEP]
                feed_input_mask.append([1] * len(feed_s1[i]))
                feed_segment_ids.append([0] * len(feed_s1[i]))

            feed_dict = {
                self.s1: utils.pad_sequences(feed_s1, padding='post'),
                self.s1_input_mask: utils.pad_sequences(feed_input_mask, padding='post'),
                self.s1_segment_ids: utils.pad_sequences(feed_segment_ids, padding='post'),
            }
            feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.

            if mode == 'infer':
                return feed_dict

            if mode in ['train', 'dev']:
                feed_ner_label = [self.bmeo2ids(label, bmeo2id) for label in batch_ner_label]
                feed_ner_label = [[bmeo2id['O']] + ele[:bert_max_length-2] + [bmeo2id['O']] for ele in feed_ner_label]
                feed_ner_label = utils.pad_sequences(feed_ner_label, padding='post')

                feed_dict[self.ner_label] = feed_ner_label
                return feed_dict

            raise ValueError(f'mode type {mode} not support')

        # 根据bert的分词器修改sent2ids的方法
        def _sent2ids(cls, sent, word2id, max_word_len = None):
            # sent 已分好词 ' '隔开
            words = sent.replace(' ', '')  # 去除已有的分词
            words = bert_tokenizer.tokenize(words)

            token_ids = [word2id.get(word, word2id['<unk>']) for word in words]
            if max_word_len:
                token_ids = token_ids[:max_word_len]
            return token_ids  # [len]

        # 替换实例方法
        import types
        self.create_feed_dict_from_features = types.MethodType(_create_feed_dict_from_features, self)
        self.create_feed_dict_from_raw = types.MethodType(_create_feed_dict_from_raw, self)
        type(self).sent2ids = types.MethodType(_sent2ids, type(self))
        print('提醒：由于该模型重写了sent2ids等函数，故生成tfrecord时check的过程中，原始分词后的字符串和转成的id可能表面上不一致，以id为准')

        bert = BERT(bert_model_dir=bert_model_dir,
                    is_training=True,
                    input_ids=self.s1,
                    input_mask=self.s1_input_mask,
                    segment_ids=self.s1_segment_ids,
                    )

        s1_mask = tf.cast(tf.sign(self.s1), tf.float32)  # [batch,len] 1 for nonpad - 0 for pad
        s1_seqlen = tf.cast(tf.reduce_sum(tf.sign(tf.abs(self.s1)), axis=1), tf.int32)  # [batch]

        bert_seq_output = bert.get_sequence_output()
        bert_seq_output = tf.layers.dropout(bert_seq_output, self.dropout_rate)  # dropout

        # logits
        cls_logits = tf.layers.dense(bert_seq_output, conf.label_size, activation=None, use_bias=True, name='cls_logits')  # [batch,length,label]

        # crf
        log_likelihood, transition_params = tf.contrib.crf.crf_log_likelihood(cls_logits, self.ner_label, tf.to_int32(s1_seqlen))
        crf_loss = tf.reduce_mean(-log_likelihood)

        # loss
        self.loss = crf_loss

        # crf decode
        ner_pred, ner_prob = tf.contrib.crf.crf_decode(cls_logits, transition_params, tf.to_int32(s1_seqlen))

        self.ner_prob = tf.identity(ner_prob, name='ner_prob')  # [batch]
        self.ner_pred = tf.identity(ner_pred, name='ner_pred')  # [batch,len]

        with tf.name_scope('accuracy'):
            ner_acc = tf.cast(tf.equal(self.ner_pred, self.ner_label), tf.float32) * s1_mask  # [batch,len]
            ner_acc = tf.reduce_sum(ner_acc, axis=-1)  # [batch]
            ner_acc = ner_acc / tf.cast(s1_seqlen, tf.float32)  # [batch]
            self.accuracy = tf.reduce_mean(ner_acc, axis=0)  # scalar

        self.global_step = tf.train.get_or_create_global_step()
        self.optimizer = tf.train.AdamOptimizer(learning_rate=conf.lr)
        self.train_op = self.optimizer.minimize(self.loss, global_step=self.global_step)

    @classmethod
    def sent2ids(cls, sent, word2id, max_word_len=None):
        # sent 已分好词 ' '隔开
        # 形成batch时才动态补齐长度
        words = sent.split(' ')
        token_ids = [word2id.get(word, word2id['<unk>']) for word in words]
        if max_word_len:
            token_ids = token_ids[:max_word_len]
        return token_ids  # [len]

    @classmethod
    def bmeo2ids(cls, ner_labels, bmeo2id, max_word_len=None):
        # ner_labels 已分好词 ' '隔开
        # 形成batch时才动态补齐长度
        labels = ner_labels.split(' ')
        token_ids = [bmeo2id.get(label, bmeo2id['<unk>']) for label in labels]
        if max_word_len:
            token_ids = token_ids[:max_word_len]
        return token_ids  # [len]

    def create_feed_dict_from_data(self, data, ids, mode='train'):
        # data:数据已经转为id, data不同字段保存该段字段全量数据
        batch_s1 = [data['s1'][i] for i in ids]
        batch_ner_label = [data['ner_label'][i] for i in ids]
        if len(set([len(e) for e in batch_s1])) != 1:  # 长度不等
            batch_s1 = utils.pad_sequences(batch_s1, padding='post')
        if len(set([len(e) for e in batch_ner_label])) != 1:  # 长度不等
            batch_ner_label = utils.pad_sequences(batch_ner_label, padding='post')
        feed_dict = {
            self.s1: batch_s1,
            self.ner_label: batch_ner_label,
        }
        if mode == 'train': feed_dict['num'] = len(batch_s1)
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_features(self, features, mode='train'):
        # feature:tfrecord数据的example, 每个features的不同字段包括该字段一个batch数据
        feed_dict = {
            self.s1: features['s1'],
            self.ner_label: features['ner_label'],
        }
        if mode == 'train': feed_dict['num'] = len(features['s1'])
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.
        return feed_dict

    def create_feed_dict_from_raw(self, batch_s1, batch_ner_label, token2id_dct, mode='infer'):
        char2id = token2id_dct['char2id']
        bmeo2id = token2id_dct['bmeo2id']

        feed_s1 = [self.sent2ids(s1, char2id) for s1 in batch_s1]

        feed_dict = {
            self.s1: utils.pad_sequences(feed_s1, padding='post'),
        }
        feed_dict[self.dropout_rate] = conf.dropout_rate if mode == 'train' else 0.

        if mode == 'infer':
            return feed_dict

        if mode in ['train', 'dev']:
            assert batch_ner_label, 'batch_ner_label should not be None when mode is train or dev'
            feed_ner_label = [self.bmeo2ids(label, bmeo2id) for label in batch_ner_label]
            feed_dict[self.ner_label] = utils.pad_sequences(feed_ner_label, padding='post')
            return feed_dict

        raise ValueError(f'mode type {mode} not support')

    @classmethod
    def generate_data(cls, file, token2id_dct):
        char2id = token2id_dct['char2id']
        bmeo2id = token2id_dct['bmeo2id']
        data = {
            's1': [],
            'ner_label': []
        }
        with open(file, 'r', encoding='U8') as f:
            for i, line in enumerate(f):
                item = line.strip().split('\t')
                s1 = item[0]
                bmeo = item[1]
                s1_ids = cls.sent2ids(s1, char2id, max_word_len=100)
                bmeo_ids = cls.bmeo2ids(bmeo, bmeo2id, max_word_len=100)
                if i < 5:  # check
                    print(f'check {i}:')
                    print(f'{s1} -> {s1_ids}')
                    print(f'{bmeo} -> {bmeo_ids}')
                data['s1'].append(s1_ids)
                data['ner_label'].append(bmeo_ids)
        data['num_data'] = len(data['s1'])
        return data

    @classmethod
    def generate_tfrecord(cls, file, token2id_dct, tfrecord_file):
        from qiznlp.common.tfrecord_utils import items2tfrecord
        char2id = token2id_dct['char2id']
        bmeo2id = token2id_dct['bmeo2id']

        def items_gen():
            with open(file, 'r', encoding='U8') as f:
                for i, line in enumerate(f):
                    item = line.strip().split('\t')
                    if len(item) != 2:
                        print(repr(line))
                        continue
                    try:
                        s1 = item[0]
                        bmeo = item[1]
                        s1_ids = cls.sent2ids(s1, char2id, max_word_len=100)
                        bmeo_ids = cls.bmeo2ids(bmeo, bmeo2id, max_word_len=100)
                        if i < 5:  # check
                            print(f'check {i}:')
                            print(f'{s1} -> {s1_ids}')
                            print(f'{bmeo} -> {bmeo_ids}')
                        d = {
                            's1': s1_ids,
                            'ner_label': bmeo_ids,
                        }
                        yield d
                    except Exception as e:
                        print('Exception occur in items_gen()!\n', e)
                        continue

        count = items2tfrecord(items_gen(), tfrecord_file)
        return count

    @classmethod
    def load_tfrecord(cls, tfrecord_file, batch_size=128, index=None, shard=None):
        from qiznlp.common.tfrecord_utils import tfrecord2dataset
        if not os.path.exists(tfrecord_file):
            return None, None
        feat_dct = {
            's1': tf.VarLenFeature(tf.int64),
            'ner_label': tf.VarLenFeature(tf.int64),
        }
        dataset, count = tfrecord2dataset(tfrecord_file, feat_dct, batch_size=batch_size, auto_pad=True, index=index, shard=shard)
        return dataset, count

    def get_signature_export_model(self):
        inputs_dct = {
            's1': self.s1,
            'dropout_rate': self.dropout_rate,
        }
        outputs_dct = {
            'ner_pred': self.ner_pred,
            'ner_prob': self.ner_prob,
        }
        return inputs_dct, outputs_dct

    @classmethod
    def get_signature_load_pbmodel(cls):
        inputs_lst = ['s1', 'dropout_rate']
        outputs_lst = ['ner_pred', 'ner_prob']
        return inputs_lst, outputs_lst

    @classmethod
    def from_pbmodel(cls, pbmodel_dir, sess):
        meta_graph_def = tf.saved_model.loader.load(sess, [tf.saved_model.tag_constants.SERVING], pbmodel_dir)  # 从pb模型载入graph和variable,绑定到sess
        signature = meta_graph_def.signature_def  # 取出signature_def
        default_key = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # or 'chitchat_predict' 签名
        inputs_lst, output_lst = cls.get_signature_load_pbmodel()  # 类方法
        pb_dict = {}
        for k in inputs_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].inputs[k].name)  # 从signature中获取输入输出的tensor_name,并从graph中取出
        for k in output_lst:
            pb_dict[k] = sess.graph.get_tensor_by_name(signature[default_key].outputs[k].name)
        model = cls(build_graph=False)  # 里面不再构造图
        for k, v in pb_dict.items():
            setattr(model, k, v)  # 绑定必要的输入输出到实例
        return model

    @classmethod
    def from_ckpt_meta(cls, ckpt_name, sess, graph):
        with graph.as_default():
            saver = tf.train.import_meta_graph(ckpt_name + '.meta', clear_devices=True)
            sess.run(tf.global_variables_initializer())

        model = cls(build_graph=False)  # 里面不再构造图
        # 绑定必要的输入输出到实例
        model.s1 = graph.get_tensor_by_name('s1:0')
        model.dropout_rate = graph.get_tensor_by_name('dropout_rate:0')
        model.ner_prob = graph.get_tensor_by_name('ner_prob:0')
        model.ner_pred = graph.get_tensor_by_name('ner_pred:0')

        saver.restore(sess, ckpt_name)
        print(f':: restore success! {ckpt_name}')
        return model, saver
